#!/usr/bin/python3
#If settings need to be changed, suggested location is stored on /mnt/usb{1,4}/integrity_check, used from ~/integrity_check (see read_keys.sh)
import os.path
import glob
import fnmatch
import subprocess
import logging
import re
import datetime
import hashlib
import collections
import itertools
import tempfile
import gc
import sys
import py_compile
import importlib.util
import json

if os.path.exists('/mnt/hd0'):
    target_root='/mnt/hd0/'#the trailing slash is required
else:
    target_root='/'
checker_root='/'
include_md5lists=False#from /var/lib/dpkg/info - these lists could themselves be tampered with and use a low-security hash, but are better than nothing for packages without a cached .deb
check_pyc=False#uselessly noisy, as .pyc files aren't reproducible (source timestamp, hash randomization, order of input files): https://www.python.org/dev/peps/pep-0552/
archive_keyring=checker_root+'/usr/share/keyrings/debian-archive-keyring.gpg'
if len(sys.argv)>2:
    raise ValueError("too many arguments")
if len(sys.argv)==2:
    settings_path=sys.argv[1]
    if not os.path.exists(settings_path+'/integrity_check_settings.json'):
        raise OSError("if given, argument must be a directory containing settings files")
else:
    settings_path=os.path.dirname(sys.argv[0])
with open(settings_path+'/integrity_check_settings.json','r') as settings_file:
    settings=json.load(settings_file)
paths_ignore_extra_full=[(target_root+p).replace('//','/') for p in settings['paths_ignore_extra']]
patterns_ignore_extra_full=[(target_root+p).replace('//','/') for p in settings['patterns_ignore_extra']]

def file_sha256(fname):
    with open(fname,'rb') as fd:
        hash1=hashlib.sha256()
        block='nonempty'
        while len(block)>0:
            block=fd.read(10**7)#don't run out of memory on very large files
            hash1.update(block)
        return hash1.hexdigest()

def file_md5(fname):
    with open(fname,'rb') as fd:
        hash1=hashlib.md5()
        block='nonempty'
        while len(block)>0:
            block=fd.read(10**7)#don't run out of memory on very large files
            hash1.update(block)
        return hash1.hexdigest()

def make_print_safe(s):#prevent terminal control characters getting into the output
    s=str(s)
    return ''.join(('?' if not (str.isprintable(c) or c=='\n') else c) for c in s)

logging.getLogger().addHandler(logging.FileHandler(settings['logfile_name'],encoding='UTF-8'))
#logging.getLogger().addHandler(logging.StreamHandler())
print("target={} checker={} settings path={}; output will be placed in {}".format(target_root,checker_root,make_print_safe(settings_path),make_print_safe(settings['logfile_name'])))
logging.warning("started at {} target={} checker={} settings path={} log file={}".format(datetime.datetime.now(),target_root,checker_root,make_print_safe(settings_path),make_print_safe(settings['logfile_name'])))
logging.warning("to check that the scan actually runs, we deliberately don't suppress all expected warnings")
release_files=glob.glob(target_root+'/var/lib/apt/lists/*Release')
packageindex_sha256={}
for f in release_files:
    fs=f+'.gpg'
    fbase=f.rsplit('_',1)[0]
    if subprocess.run([checker_root+"/usr/bin/gpgv","--keyring",archive_keyring,"--weak-digest","SHA1"]+([fs] if os.path.exists(fs) else [])+[f],stderr=subprocess.PIPE).returncode==0:
        with open(f,'r') as fd:
            is_sha256=False
            for line in fd:
                if is_sha256:
                    r=re.fullmatch(" ([0-9a-f]{64})\s+[0-9]+\s(.*)\n",line)
                    if r:
                        pfhash,pfname=r.groups()
                        packageindex_sha256[fbase+'_'+pfname.replace('/','_')]=pfhash
                    else:
                        is_sha256=False
                if line=='SHA256:\n':
                    is_sha256=True
                if line.startswith('Valid-Until:'):
                    expiry=datetime.datetime.strptime(line,'Valid-Until: %a, %d %b %Y %H:%M:%S %Z\n')
                    if expiry<datetime.datetime.now():
                        logging.warning("Release file outdated: "+f+" "+line)
    else:
        logging.error("Release file bad signature: "+make_print_safe(f))
packageindex_files=glob.glob(target_root+'/var/lib/apt/lists/*Packages')
#print(release_files,packageindex_files,packageindex_sha256)
deb_files=glob.glob((target_root+'/var/cache/apt/archives/*.deb').replace('//','/'))
verified_deb_files=[]
available_packages=collections.defaultdict(list)
for f in packageindex_files:
    fhash=file_sha256(f)
    if f in packageindex_sha256:
        if packageindex_sha256[f]==fhash:
            with open(f,'r') as fd:
                current_package={}
                for line in itertools.chain(fd,['\n']):#ensure the last package gets processed
                    for fieldname in ["Package","Version","Filename","SHA256","Architecture","Multi-Arch","Installed-Size"]:
                        r=re.fullmatch(fieldname+': (.*)\n',line)
                        if r:
                            current_package[fieldname]=r.group(1)
                    if line=='\n' and current_package:#empty line=end of package
                        current_package['index_file']=f
                        deb_path=(target_root+'/var/cache/apt/archives/'+current_package['Package']+'_'+current_package['Version'].replace(':','%3a')+'_'+current_package['Architecture']+'.deb').replace('//','/')
                        if os.path.exists(deb_path):
                            debhash=file_sha256(deb_path)
                            if debhash==current_package['SHA256']:
                                verified_deb_files.append(deb_path)
                                current_package['cache_file']=deb_path
                            else:
                                logging.error(".deb file bad hash: {} expected {} found {}".format(deb_path,current_package['SHA256'],debhash))
                        available_packages[current_package['Package']].append(current_package)
                        current_package={}
        else:
            logging.error("Packages file bad hash: {} expected {} found {}".format(f,packageindex_sha256[f],fhash))
    else:
        logging.warning("Packages file not listed in Release files - ignoring: "+make_print_safe(f))
#print(verified_deb_files,available_packages)
for deb_file in deb_files:
    if deb_file not in verified_deb_files:
        logging.warning(".deb {} not found in Packages file".format(make_print_safe(deb_file)))#as this will warn for old versions, consider ignoring it
gc.collect()
installed_packages=[]
confonly_packages=[]
with open(target_root+'/var/lib/dpkg/status','r') as fd:
    current_package={}
    for line in itertools.chain(fd,['\n']):#ensure the last package gets processed
        for fieldname in ["Package","Version","Status","Architecture","Multi-Arch"]:
            r=re.fullmatch(fieldname+': (.*)\n',line)
            if r:
                current_package[fieldname]=r.group(1)
        if line=='\n':
            if current_package.get("Status",'')=='install ok installed':#empty line=end of package
                installed_packages.append(current_package)
            elif current_package.get("Status",'')=='deinstall ok config-files':
                confonly_packages.append(current_package)
            elif current_package!={}:
                logging.warning("Non-fully-installed package {}".format(make_print_safe(current_package)))
            current_package={}
if confonly_packages:
    logging.warning("Removed-but-not-purged packages (may cause spurious 'file in no package' warnings): {}".format(' '.join(make_print_safe(p.get('Package','')) for p in confonly_packages)))
installed_packages_with_deb_files=[]
packages_md5lists=[]
packages_missing_debs=[]
for current_package in installed_packages:
    known_versions=[p['Version'] for p in available_packages[current_package['Package']]]
    if len(known_versions)==0:
        logging.warning("Package {} not listed in Packages files".format(make_print_safe(current_package['Package'])))
    elif current_package['Version'] not in known_versions:
        logging.warning("Package {} has unknown version {} (expected one of {})".format(current_package['Package'],make_print_safe(current_package['Version'],known_versions)))
    elif len(known_versions)>1:
        for v in known_versions:
            if subprocess.run([checker_root+"/usr/bin/dpkg","--compare-versions",current_package['Version'],"lt",v]).returncode==0:
                logging.warning("Package {} is outdated: {} installed, {} available".format(current_package['Package'],current_package['Version'],v))
    deb_entry=[p for p in available_packages[current_package['Package']] if (p['Version']==current_package['Version'] and p['Architecture']==current_package['Architecture'] and 'cache_file' in p.keys())]
    if len(deb_entry)==0:
        packages_missing_debs.append(current_package['Package'])
        if include_md5lists:
            packages_md5lists.append(target_root+'/var/lib/dpkg/info/'+current_package['Package']+((':'+current_package['Architecture']) if current_package.get('Multi-Arch','')=='same' else ''))
    elif len(set(p['SHA256'] for p in deb_entry))>1:#having the same package in X and X-updates is not an error, but having different package files with the same version+arch is
        logging.error("Duplicate packages {}".format(deb_entry))
    else:
        installed_packages_with_deb_files.append(deb_entry[0])
if(packages_missing_debs):
    logging.warning("To get missing .deb files: apt-get download "+make_print_safe(" ".join(packages_missing_debs)))
del available_packages
installed_packages_with_deb_files=sorted(installed_packages_with_deb_files,key=lambda p:-int(p.get('Installed-Size',10000)))#process large packages first to minimize risk of running out of memory on a liveCD ramdisk
matched_files=set()
expected_files=collections.defaultdict(list)
for current_package in installed_packages_with_deb_files:
    deb_file=current_package['cache_file']
    gc.collect()
    with tempfile.TemporaryDirectory() as tmpdir1:
        if subprocess.run([checker_root+"/usr/bin/dpkg-deb","--raw-extract",deb_file,tmpdir1]).returncode!=0:
            logging.error("Failed to extract .deb file "+deb_file)
        for dirname,subdirs,files in os.walk(tmpdir1):
            if dirname==tmpdir1 and 'DEBIAN' in subdirs:
                subdirs.remove('DEBIAN')
            for fbasename in files:
                pkg_fname=os.path.join(dirname,fbasename)
                fs_fname=(target_root+'/'+os.path.relpath(pkg_fname,start=tmpdir1)).replace('//','/')
                #we omit the directory, and check for match here rather than later, to save memory: a full set of hashes can be GBs
                current_file={'from':os.path.basename(deb_file)}
                expected_hash=None
                expected_target=None
                if os.path.islink(pkg_fname):
                    expected_target=os.readlink(pkg_fname)
                    if os.path.isabs(expected_target):#absolute link
                        expected_target=(target_root+expected_target).replace('//','/')
                    else:#relative link
                        expected_target=(target_root+os.path.relpath(os.path.join(dirname,expected_target),start=tmpdir1)).replace('//','/')
                else:
                    expected_hash=file_sha256(pkg_fname)
                    expected_copies=[]
                    if fs_fname==(target_root+'/usr/share/grub/unicode.pf2').replace('//','/'):
                        expected_copies=[(target_root+'/boot/grub/unicode.pf2').replace('//','/'),(target_root+'/boot/grub/fonts/unicode.pf2').replace('//','/')]
                    re1=re.fullmatch((target_root+'/usr/share/locale/([^/]+)/LC_MESSAGES/grub.mo').replace('//','/'),fs_fname)
                    if re1:
                        expected_copies=[(target_root+'/boot/grub/locale/'+re1.group(1)+'.mo').replace('//','/')]
                    re1=re.fullmatch((target_root+'/usr/lib/grub/([^-/]+-efi/[^/]*)').replace('//','/'),fs_fname)
                    if re1 and not fs_fname.endswith('config.h')  and not fs_fname.endswith('kernel.img'):
                        expected_copies=[(target_root+'/boot/grub/'+re1.group(1)).replace('//','/')]
                    for fname2 in expected_copies:
                        expected_files[fname2].append({'from':os.path.basename(deb_file),'expected_sha256':expected_hash})
                        if not os.path.lexists(fname2):
                            logging.error("File {} copied from package {} missing".format(fname2,deb_file))
                if not os.path.lexists(fs_fname):
                    logging.error("File {} from package {} missing".format(fs_fname,deb_file))
                else:
                    try:
                        if os.path.islink(fs_fname):
                            actual_target=os.readlink(fs_fname)
                            if os.path.isabs(actual_target):#absolute link
                                actual_target=(target_root+actual_target).replace('//','/')
                            else:#relative link
                                actual_target=os.path.normpath(os.path.join(os.path.dirname(fs_fname),actual_target))
                            if expected_target is not None and expected_target==actual_target:
                                matched_files.add(fs_fname)
                            else:
                                current_file['actual_target']=make_print_safe(actual_target)
                                current_file['expected_target']=expected_target
                                expected_files[fs_fname].append(current_file)
                        else:
                            actual_hash=file_sha256(fs_fname)
                            if expected_hash is not None and expected_hash==actual_hash:
                                matched_files.add(fs_fname)
                            else:
                                current_file['actual_hash']=actual_hash
                                current_file['expected_sha256']=expected_hash
                                expected_files[fs_fname].append(current_file)
                    except OSError as err:
                        logging.warning("Can't read file {}: {}".format(fs_fname,make_print_safe(err)))
                        current_file['unreadable']=True
                        expected_files[fs_fname].append(current_file)
        #check control files
        controlfile_basename=(target_root+'/var/lib/dpkg/info/'+current_package['Package']+((':'+current_package['Architecture']) if current_package.get('Multi-Arch','')=='same' else '')+'.').replace('//','/')
        deb_control_files=glob.glob(tmpdir1+'/DEBIAN/*')
        if tmpdir1+'/DEBIAN/control' in deb_control_files:
            deb_control_files.remove(tmpdir1+'/DEBIAN/control')
        for fname in deb_control_files:
            fname2=controlfile_basename+os.path.basename(fname)
            if os.path.exists(fname2):
                if file_sha256(fname)!=file_sha256(fname2):
                    logging.error("Control file bad hash: {} {} != {}".format(fname,file_sha256(fname),file_sha256(fname2)))
            else:
                logging.error("Control file {} missing".format(fname2))
        fs_control_files=glob.glob(controlfile_basename+'*')
        fs_control_files=[f for f in fs_control_files if f.count('.')==controlfile_basename.count('.')]#stops e.g. "python3.5.md5sums" matching "python3.*"
        if controlfile_basename+'list' in fs_control_files:
            fs_control_files.remove(controlfile_basename+'list')
        if controlfile_basename+'md5sums' in fs_control_files and tmpdir1+'/DEBIAN/md5sums' not in deb_control_files:
            with open(controlfile_basename+'md5sums','r') as fd:
                if fd.read(100)=='':#empty metapackages may have no md5sums in the package but an empty one in dpkg/info
                    fs_control_files.remove(controlfile_basename+'md5sums')
        if len(fs_control_files)>len(deb_control_files):
            logging.error("Extra control file(s) {} {}".format(deb_control_files,make_print_safe(fs_control_files)))
for md5_file in packages_md5lists:
    with open(md5_file+'.md5sums','r') as fd:
        for line in fd:
            r=re.fullmatch("([0-9a-f]{32})\s+(.*)\n",line)
            if r:
                fhash,fname=r.groups()
                fname=(target_root+fname).replace('//','/')
                current_file={'from':md5_file,'md5':fhash}
                expected_files[fname].append(current_file)
                if not os.path.lexists(fname):
                    logging.error("File {} from package {} missing".format(fname,md5_file))
    if os.path.exists(md5_file+'.conffiles'):
        with open(md5_file+'.conffiles','r') as fd:
            for line in fd:
                r=re.fullmatch("(.*)\n",line)
                if r:
                    fname,=r.groups()
                    fname=(target_root+fname).replace('//','/')
                    current_file={'from':md5_file,'conffile_nohash':True}
                    if len(expected_files[fname])==0:
                        expected_files[fname].append(current_file)
                    if not os.path.lexists(fname):
                        logging.error("File {} from package {} missing".format(fname,md5_file))
    with open(md5_file+'.list','r') as fd:
        for line in fd:
            r=re.fullmatch("(.*)\n",line)
            if r:
                fname,=r.groups()
                fname=(target_root+fname).replace('//','/')
                if len(expected_files[fname])==0:#no md5 listed = directory or symlink
                    current_file={'from':md5_file,'symlink_target':None}
                    expected_files[fname].append(current_file)
                if not os.path.lexists(fname):
                    logging.error("File {} from package {} missing".format(fname,md5_file))
for fname in settings['localmod_files'].keys():
    fs_fname=(target_root+fname).replace('//','/')
    expected_files[fs_fname]=[{'from':'localmod','expected_sha256':settings['localmod_files'][fname]}]
    if fs_fname in matched_files:
        matched_files.remove(fs_fname)#localmod overrides packaged version
gc.collect()
expected_pycfiles={}
for fname in expected_files.keys():
    if len(expected_files[fname])!=1:
        logging.warning("File {} in multiple packages {}".format(fname,expected_files[fname]))
    match=False
    failed_read=False
    if fname in matched_files:
        match=True
    for current_file in expected_files[fname]:
        if current_file.get('unreadable',False):
            failed_read=True
        elif 'expected_sha256' in current_file.keys():
            try:
                if file_sha256(fname)==current_file['expected_sha256']:
                    match=True
            except OSError as err:
                logging.warning("Can't read file {}: {}".format(fname,make_print_safe(err)))
                failed_read=True
        elif include_md5lists and 'md5' in current_file.keys():
            try:
                if file_md5(fname)==current_file['md5']:
                    match=True
            except OSError as err:
                logging.warning("Can't read file {}: {}".format(fname,make_print_safe(err)))
                failed_read=True
        elif include_md5lists and 'symlink_target' in current_file.keys() and current_file['symlink_target'] is None:
            if os.path.islink(fname) or not os.path.isfile(fname):
                match=True
        elif include_md5lists and current_file.get('conffile_nohash',False):
            match=True
    if not (match or failed_read):
        logging.error("File bad hash: {} expected {}".format(fname,expected_files[fname]))
    if match and check_pyc and os.path.splitext(fname)[1]=='.py':
        try:
            pyc_name=importlib.util.cache_from_source(fname)
            with tempfile.TemporaryDirectory() as tmpdir2:
                py_compile.compile(fname,tmpdir2+'/check.pyc',doraise=True)
                pyc_hash=file_sha256(tmpdir2+'/check.pyc')
                expected_pycfiles[pyc_name]=pyc_hash
        except py_compile.PyCompileError as err:
            expected_pycfiles[pyc_name]=None
            logging.warning("{} .pyc compile error {}".format(fname,err))
gc.collect()
pycfile_count=0
for dirname,subdirs,files in os.walk(target_root,topdown=True):
    for dbasename in subdirs.copy():#copy because we're modifying the list, which might break iterators
        if os.path.join(dirname,dbasename) in paths_ignore_extra_full:
            subdirs.remove(dbasename)
    for fbasename in files:
        fname=os.path.join(dirname,fbasename)
        f_expected=expected_files[fname]
        if len(f_expected)==0 and (not any(fnmatch.fnmatch(fname,p) for p in patterns_ignore_extra_full)) and fname not in matched_files:
            if os.path.islink(fname):
                ftarget=fname
                while os.path.islink(ftarget):
                    ftarget2=os.readlink(ftarget)
                    if os.path.isabs(ftarget2):#absolute link
                        ftarget=(target_root+ftarget2).replace('//','/')
                    else:#relative link
                        ftarget=os.path.normpath(os.path.join(os.path.dirname(ftarget),ftarget2))
                #too noisy logging.warning("Symbolic link {} to {}".format(make_print_safe(fname),make_print_safe(os.readlink(fname))))
                if os.path.exists(fname) and len(expected_files[ftarget])==0 and ftarget not in matched_files:
                    logging.warning("Symbolic link {} to non-package file {}".format(make_print_safe(fname),make_print_safe(os.readlink(fname))))
            elif os.path.isfile(fname):
                try:
                    fhash=file_sha256(fname)
                    if os.path.splitext(fname)[1]=='.pyc':
                        pycfile_count=pycfile_count+1
                        if check_pyc:
                            if fname in expected_pycfiles:
                                if expected_pycfiles[fname] is not None and fhash!=expected_pycfiles[fname]:
                                    logging.warning(".pyc file {} bad hash: expected {} found {}".format(make_print_safe(fname),expected_pycfiles[fname],fhash))
                            else:
                                if not fname.startswith((target_root+'/usr/lib/python2.7/').replace('//','/')):
                                    logging.warning("Unexpected .pyc file {}, sha256={}".format(make_print_safe(fname),fhash))
                    else:
                        logging.warning("File {} belongs to no package, sha256={}".format(make_print_safe(fname),fhash))
                except OSError as err:
                    logging.warning("File {} belongs to no package, can't be read: {}".format(make_print_safe(fname),make_print_safe(err)))
            else:
                logging.warning("Non-regular file {}".format(make_print_safe(fname)))
if not check_pyc:
    logging.warning("{} .pyc files not checked".format(pycfile_count))
if os.path.exists(settings_path+"/expected_grub_cfg.txt"):
    logging.warning("grub.cfg check result (blank=OK, note that differences are expected if the Linux ABI number changes: {}".format(make_print_safe(str(subprocess.run([checker_root+"/usr/bin/diff",os.path.dirname(sys.argv[0])+"/expected_grub_cfg.txt",target_root+"/boot/grub/grub.cfg"],stdout=subprocess.PIPE).stdout,encoding='UTF-8'))))
else:
    logging.warning("expected_grub_cfg.txt not found")
logging.warning("finished at {} log file={}".format(datetime.datetime.now(),make_print_safe(settings['logfile_name'])))
